import json
from typing import Any, Dict, List, Optional, Union

from common.backend.models.base import ConversationParams, MediaSummary, RetrievalSummaryJson
from common.llm_assist.logging import logger
from common.solutions.chains.no_retrieval_chain import NoRetrievalChain
from dataiku.langchain.dku_llm import DKULLM
from portal.backend.models import (
    AgentDetails,
    AgentGeneratedAnswer,
    AgentQuery,
    AgentsSelection,
    PortalLLMContext,
    PortalSource,
    ToolCall,
)
from portal.backend.utils.sources_utils import process_agents_sources, process_answers_sources


class QueryResolverChain(NoRetrievalChain):
    def __init__(
        self,
        llm: DKULLM,
        generated_queries_answers: List[AgentGeneratedAnswer],
        agents: List[AgentDetails],
        agents_queries: List[AgentQuery],
        query_builder_justification: str,
        chat_has_media: bool = False,
        agents_files_uploads: Optional[Dict[str, Dict[str, MediaSummary]]] = None,
        uploaded_files: Optional[List[MediaSummary]] = None,
    ):
        super().__init__(llm=llm, chat_has_media=chat_has_media)
        self.__act_like_prompt = """
        # Role and Guidelines
        You are an AI that can synthesize information from different sources and provide a final answer to the human query.
        Your role is to read the initial human query and any answers generated by different AIs to answer parts of the query or all and provide a final answer.
        """
        self.__system_prompt = f"""
        Given the initial human query, any possible generated answers, and the context of the conversation, provide a final answer to the initial human query.
        Don't change the answer generated by other AIs. If you need to provide a different answer, provide it in a new answer and mention that you are providing a different answer.
        Here is additional information that might help you about the available external agents:
        {json.dumps(agents).replace("{", "{{").replace("}", "}}")}
        """
        self.agents = agents
        self.agents_files_uploads = agents_files_uploads
        self.agents_queries = agents_queries
        self.uploaded_files = uploaded_files
        self.generated_queries_answers = generated_queries_answers
        self.query_builder_justification = query_builder_justification

    @property
    def act_like_prompt(self) -> str:
        return self.__act_like_prompt

    @property
    def system_prompt(self) -> str:
        return self.__system_prompt

    def finalize_non_streaming(
        self,
        params: ConversationParams,
        question_context: Union[str, Dict[str, Any], List[str]],
    ) -> RetrievalSummaryJson:
        return self._get_as_json(
            generated_answer=question_context,
            uploaded_docs=self.uploaded_files,
            user_profile=params.get("user_profile", None),
            sources=self.format_sources(self.generated_queries_answers),
        )

    def finalize_streaming(
        self,
        params: ConversationParams,
        question_context: Union[str, Dict[str, Any], List[str]],
    ) -> RetrievalSummaryJson:
        return self._get_as_json(
            generated_answer=question_context,
            uploaded_docs=self.uploaded_files,
            user_profile=params.get("user_profile", None),
            sources=self.format_sources(self.generated_queries_answers),
        )

    def format_sources(self, sources: List[AgentGeneratedAnswer]) -> List[PortalSource]:
        formatted_sources: List[PortalSource] = []
        logger.debug(f"Formatting sources")
        # TODO refactor with agents_orchestrator
        for source in sources:
            is_agent = source.get("agent_id", "").startswith("agent")
            items: List[ToolCall] = []
            if source.get("sources"):
                if is_agent:
                    items = process_agents_sources(source["sources"])
                else:
                    used_tool = source.get("used_tool")
                    items = process_answers_sources(source["sources"], used_tool)
            formatted_sources.append(
                {
                    "name": source["agent_name"] or "No agent name provided",
                    "id": source["agent_id"],
                    "type": "agent" if is_agent else "answers webapp",
                    "items": items,
                    "answer": source.get("ai_answer", ""),
                }
            )
        return formatted_sources

    def build_llm_context(
        self,
        user_profile: Optional[Dict[str, Any]] = None,
        uploaded_docs: Optional[List[MediaSummary]] = None,
    ) -> PortalLLMContext:
        llm_context: PortalLLMContext = {}
        if user_profile:
            llm_context["user_profile"] = user_profile
        if uploaded_docs:
            llm_context["uploaded_docs"] = [
                {
                    "original_file_name": uploaded_doc.get("original_file_name"),
                    "metadata_path": uploaded_doc.get("metadata_path", ""),
                }
                for uploaded_doc in uploaded_docs
            ]
        if self.agents_files_uploads:
            llm_context["agents_files_uploads"] = self.agents_files_uploads
        llm_context["agents_selection"] = AgentsSelection(
            calls=[
                AgentQuery(
                    query=agent_query_answer.get("sub_query", "no query provided"),
                    agent_id=agent_query_answer.get("agent_id", "no agent id provided"),
                )
                for agent_query_answer in self.generated_queries_answers
            ],
            justification=self.query_builder_justification,
        )
        return llm_context

    def _get_as_json(
        self,
        generated_answer: Union[str, Dict[str, Any], List[str]],
        user_profile: Optional[Dict[str, Any]] = None,
        uploaded_docs: Optional[List[MediaSummary]] = None,
        sources: Optional[List[PortalSource]] = [],
    ) -> RetrievalSummaryJson:
        resp: RetrievalSummaryJson = {}
        answer = ""
        llm_context: PortalLLMContext = self.build_llm_context(user_profile=user_profile, uploaded_docs=uploaded_docs)
        gen_user_profile = user_profile
        if isinstance(generated_answer, str):
            answer = generated_answer
        if isinstance(generated_answer, dict):
            answer = generated_answer.get("answer", "")
            gen_user_profile = generated_answer.get("user_profile", user_profile)
            if generated_answer.get("images", None):
                resp["generated_images"] = generated_answer["images"]
        else:
            logger.error(f"Generated answer is not a dictionary: {generated_answer}")
            return resp
        resp.update(
            {
                "answer": answer,
                "sources": sources or [],
                "filters": None,
                "llm_context": llm_context,
                "user_profile": gen_user_profile or {},
            }
        )
        return resp
